"""

Revision ID: 6e324523861c
Revises: ea815a161929
Create Date: 2025-06-21 17:08:08.289299

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa

import sqlmodel

# revision identifiers, used by Alembic.
revision: str = "6e324523861c"
down_revision: Union[str, None] = "ea815a161929"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("subject", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "symbol_without_suffix",
                sqlmodel.sql.sqltypes.AutoString(),
                nullable=True,
            )
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("subject", schema=None) as batch_op:
        batch_op.drop_column("symbol_without_suffix")

    # ### end Alembic commands ###



"""

Revision ID: ea815a161929
Revises: 
Create Date: 2025-05-31 17:11:37.371121

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import sqlmodel

# revision identifiers, used by Alembic.
revision: str = "ea815a161929"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "subject",
        sa.Column("reason", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("high_price_target", sa.Float(), nullable=True),
        sa.Column("low_price_target", sa.Float(), nullable=True),
        sa.Column("fair_value", sa.Float(), nullable=True),
        sa.Column(
            "summary_price_target", sqlmodel.sql.sqltypes.AutoString(), nullable=True
        ),
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("consensus", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("date", sa.DateTime(), nullable=False),
        sa.Column("news", sa.JSON(), nullable=True),
        sa.Column("news_summary", sa.JSON(), nullable=True),
        sa.Column("summary", sa.JSON(), nullable=True),
        sa.Column("price_target_news", sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint("symbol", "date"),
    )
    with op.batch_alter_table("subject", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_subject_consensus"), ["consensus"], unique=False
        )
        batch_op.create_index(batch_op.f("ix_subject_date"), ["date"], unique=False)
        batch_op.create_index(batch_op.f("ix_subject_symbol"), ["symbol"], unique=False)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("subject", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_subject_symbol"))
        batch_op.drop_index(batch_op.f("ix_subject_date"))
        batch_op.drop_index(batch_op.f("ix_subject_consensus"))

    op.drop_table("subject")
    # ### end Alembic commands ###



from typing import Optional

from pydantic import BaseModel


class Summary(BaseModel):
    content: str


class BaseArticle(Summary):
    url: Optional[str] = None
    source: str
    title: Optional[str] = None

    def __hash__(self) -> int:
        return hash((self.url, self.source))


class News(BaseArticle): ...


class NewsSummary(BaseArticle): ...


class PriceTargetNews(BaseArticle): ...



import json
import logging
import re
import tempfile
from abc import abstractmethod
from contextlib import contextmanager
from pathlib import Path
from time import sleep
from typing import List, Optional, Generator, Any, Callable

from bs4 import BeautifulSoup
from pydantic import BaseModel, model_validator
from selenium import webdriver
from selenium.common import NoSuchElementException
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.chrome.webdriver import WebDriver
from selenium.webdriver.common.by import By
from webdriver_manager.chrome import ChromeDriverManager

from tickermood.articles import News, PriceTargetNews
from tickermood.subject import Subject
from tickermood.types import SourceName

import yfinance as yf  # type: ignore[import-untyped]

logger = logging.getLogger(__name__)
PAGE_SOURCE_PATH = Path(__file__).parents[1] / "tests" / "sources"


def clean_url(url: str) -> str:
    return re.sub(r"[^a-zA-Z0-9]", "", url)


class SavePage(BaseModel):
    url: str
    source: str
    save: bool = False

    @model_validator(mode="after")
    def _validator(self) -> "SavePage":
        self.url = f'{clean_url(self.url.replace("html",""))}.html'
        return self


@contextmanager
def web_browser(
    url: str,
    load_strategy_none: bool = False,
    headless: bool = False,
    callback: Optional[Callable[[WebDriver], None]] = None,
) -> Generator[WebDriver, Any, None]:
    option = Options()
    option.add_argument("--disable-popup-blocking")
    option.add_argument(
        "user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) "
        "Chrome/123.0.0.0 Safari/537.36"
    )
    option.add_argument("--no-sandbox")
    option.add_argument("--disable-dev-shm-usage")
    if headless:
        option.add_argument("--headless")
    if load_strategy_none:
        option.page_load_strategy = "none"
    browser = webdriver.Chrome(
        service=Service(ChromeDriverManager().install()), options=option
    )
    browser.get(url)
    if callback:
        sleep(2)
        callback(browser)

    sleep(2)

    yield browser

    browser.quit()


@contextmanager
def soup_page(
    browser: WebDriver, save_page: Optional[SavePage] = None
) -> Generator[BeautifulSoup, Any, None]:
    with tempfile.NamedTemporaryFile(suffix=".html", delete=True) as page:
        page_source_code = browser.page_source.encode("utf-8")
        Path(page.name).write_bytes(page_source_code)
        if save_page and save_page.save:
            source_path = PAGE_SOURCE_PATH.joinpath(save_page.source)
            source_path.mkdir(parents=True, exist_ok=True)
            source_path.joinpath(clean_url(save_page.url)).write_bytes(page_source_code)

        yield BeautifulSoup(page, "html.parser")


@contextmanager
def local_html(
    url: str, load_strategy_none: bool = False, headless: bool = False
) -> Generator[str, Any, None]:
    with web_browser(
        url, load_strategy_none, headless
    ) as browser, tempfile.NamedTemporaryFile(suffix=".html", delete=True) as page:
        Path(page.name).write_bytes(browser.page_source.encode("utf-8"))
        yield "file://" + page.name


@contextmanager
def temporary_web_page(
    url: str,
    load_strategy_none: bool = False,
    headless: bool = False,
    save_page: Optional[SavePage] = None,
    callback: Optional[Callable[[WebDriver], None]] = None,
) -> Generator[BeautifulSoup, Any, None]:
    with web_browser(
        url, load_strategy_none, headless, callback=callback
    ) as browser, soup_page(browser, save_page=save_page) as soup:
        yield soup


class BaseSource(BaseModel):
    name: SourceName
    url: str
    headless: bool = False

    @classmethod
    def search_subject(cls, subject: Subject, headless: bool = False) -> Subject:
        source = cls.search(subject, headless=headless)
        if source:
            subject.news.extend(source.news())
            subject.price_target_news.extend(source.get_price_target_news())
        return subject

    @classmethod
    def search(
        cls, subject: Subject, headless: bool = False
    ) -> Optional["BaseSource"]: ...

    @abstractmethod
    def news(self) -> List[News]: ...
    @abstractmethod
    def get_price_target_news(self) -> List[PriceTargetNews]: ...


class BaseSeleniumScrapper(BaseModel): ...


class Investing(BaseSource):
    name: SourceName = "Investing"

    @classmethod
    def search(cls, subject: Subject, headless: bool = False) -> Optional["Investing"]:
        search_url = f"https://www.investing.com/search?q={subject.to_symbol_search()}"
        ticker_link = None
        save_page = SavePage(url=search_url, source="Investing", save=True)
        with temporary_web_page(
            search_url, headless=headless, save_page=save_page
        ) as soup:
            sections = soup.find_all("div", class_="searchSectionMain")
            for section in sections:
                header = section.find(class_="groupHeader")
                if header and header.get_text(strip=True) == "Quotes":
                    links = [a["href"] for a in section.find_all("a", href=True)]
                    if links:
                        ticker_link = links[0]
        if ticker_link:
            ticker_url = f"https://www.investing.com{ticker_link}"
            return cls(url=ticker_url, headless=headless)
        return None

    def news(self) -> List[News]:
        news_url = f"{self.url}-news"
        urls: List[str] = []
        articles = []
        with temporary_web_page(news_url, headless=self.headless) as soup:
            news_ = soup.find("ul", attrs={"data-test": "news-list"})
            if not news_:
                logger.warning(f"No news found at {news_url}")
                return []
            for item in news_:

                if not item.select_one(".mb-1.mt-2\\.5.flex"):  # type: ignore[union-attr]
                    links = item.find_all("a", href=True)  # type: ignore[union-attr]
                    urls.extend(list({a["href"] for a in links}))
        for url in urls:
            try:
                with temporary_web_page(url, headless=self.headless) as soup:
                    article_ = soup.find("div", class_="article_container")
                    if article_ is not None:
                        content = article_.get_text(separator=" ", strip=True)
                        articles.append(
                            News(url=url, source=self.name, content=content)
                        )
            except Exception as e:  # noqa: PERF203
                logger.warning(f"Error processing article {url}: {e}")
                continue
        return articles

    def get_price_target_news(self) -> List[PriceTargetNews]:
        consensus_url = f"{self.url}-consensus-estimates"
        with temporary_web_page(consensus_url, headless=self.headless) as soup:
            articles = soup.find_all("div", class_="mb-6")
            content = "\n\n\n".join(
                [a.get_text(separator="\n", strip=True) for a in articles]
            )
        return [PriceTargetNews(url=consensus_url, content=content, source=self.name)]


def find_cookie_banner(browser: WebDriver) -> None:
    try:
        button = browser.find_element(
            By.XPATH, "/html/body/div/div/div/div/form/div[2]/div[2]/button[1]"
        )
        button.click()
    except NoSuchElementException:
        pass


class Yahoo(BaseSource):
    name: SourceName = "Yahoo"

    @classmethod
    def search(cls, subject: Subject, headless: bool = False) -> Optional["Yahoo"]:
        return cls(
            url=subject.symbol,
            headless=headless,
        )

    def news(self) -> List[News]:
        ticker = yf.Ticker(self.url)
        urls = [
            n.get("content", {}).get("canonicalUrl", {}).get("url", "")
            for n in ticker.get_news()
        ]
        articles = []
        for url in urls:
            if not url:
                continue
            try:
                with temporary_web_page(
                    url, headless=self.headless, callback=find_cookie_banner
                ) as soup:

                    for tag in soup(
                        [
                            "script",
                            "style",
                            "noscript",
                            "svg",
                            "meta",
                            "header",
                            "footer",
                        ]
                    ):
                        tag.decompose()
                    if soup is not None:
                        content = soup.get_text(separator=" ", strip=True)
                        articles.append(
                            News(url=url, source=self.name, content=content)
                        )
            except Exception as e:
                logger.warning(f"Error processing article {url}: {e}")
                continue
        return articles

    def get_price_target_news(self) -> List[PriceTargetNews]:
        ticker = yf.Ticker(self.url)
        return [
            PriceTargetNews(
                url=self.url,
                content=json.dumps(ticker.get_analyst_price_targets()),
                source=self.name,
            )
        ]



import logging
import os
from pathlib import Path
from typing import List, Type, Annotated, Optional

import typer
from dotenv import load_dotenv
from langchain_ollama import ChatOllama
from langchain_openai import ChatOpenAI
from pydantic import BaseModel, Field

from tickermood.agent import invoke_summarize_agent
from tickermood.source import BaseSource, Investing, Yahoo
from tickermood.subject import Subject, LLM, LLMSubject
from tickermood.types import DatabaseConfig

logger = logging.getLogger(__name__)
app = typer.Typer()


class TickerMoodNews(BaseModel):
    sources: List[Type[BaseSource]] = Field(default=[Investing, Yahoo])
    subjects: List[Subject]
    headless: bool = True
    database_config: DatabaseConfig = Field(default_factory=DatabaseConfig)

    def set_database(self, database_config: Optional[DatabaseConfig] = None) -> None:
        if database_config:
            self.database_config = database_config

    @classmethod
    def from_symbols(cls, symbols: List[str]) -> "TickerMoodNews":
        subjects = [Subject(symbol=symbol) for symbol in symbols]
        return cls(subjects=subjects)

    def summarize(self, subject: Subject, llm: LLM) -> None:
        llm_subject = LLMSubject.from_subject(subject, llm)
        summarized_subject = invoke_summarize_agent(llm_subject)
        summarized_subject.save(self.database_config)

    def search(self, llm: Optional[LLM] = None) -> None:
        for subject in self.subjects:
            for source in self.sources:
                try:
                    subject = source.search_subject(  # noqa: PLW2901
                        subject, headless=self.headless
                    )
                except Exception as e:  # noqa: PERF203
                    logger.warning(
                        f"Error searching for subject {subject.symbol} in {type(source).__name__}: {e}"
                    )
                    continue
            subject.save(self.database_config)
            if llm:
                self.summarize(subject, llm)


class TickerMood(TickerMoodNews):
    llm: LLM = Field(
        default_factory=lambda: LLM(
            model_name="qwen3:4b", model_type=ChatOllama, temperature=0.0
        )
    )

    @classmethod
    def from_subjects(cls, subjects: List[Subject]) -> "TickerMood":
        return cls(subjects=subjects)

    @classmethod
    def from_symbols(cls, symbols: List[str]) -> "TickerMood":
        subjects = [Subject(symbol=symbol) for symbol in symbols]
        return cls(subjects=subjects)

    def set_llm(self, llm: LLM) -> None:
        self.llm = llm

    def run(self) -> None:
        self.search()
        self.call_agent()
        logger.info("TickerMood run completed.")

    def call_agent(self) -> None:
        if self.llm is None:
            raise ValueError("LLM must be set before calling the agent.")
        for subject in self.subjects:
            self.summarize(subject, self.llm)


def get_news(
    symbols: List[str],
    database_config: DatabaseConfig,
    headless: bool = True,
    model_name: str = "gpt-4-turbo",
) -> None:
    llm = None
    ticker_mood = TickerMoodNews.from_symbols(symbols)
    ticker_mood.set_database(database_config)
    ticker_mood.headless = headless
    if "OPENAI_API_KEY" in os.environ:
        llm = LLM(model_name=model_name, model_type=ChatOpenAI, temperature=0.0)
    ticker_mood.search(llm)


@app.command()
def run(
    symbols: Annotated[List[str], typer.Argument()],
    path: Optional[Path] = None,
    openai_api_key_path: Optional[Path] = None,
) -> None:
    ticker_mood = TickerMood.from_symbols(symbols)
    ticker_mood.set_database(DatabaseConfig(database_path=path))
    if openai_api_key_path:
        openai_api_key_path = Path(openai_api_key_path)
        if not openai_api_key_path.exists():
            raise FileNotFoundError(
                f"OpenAI API key file not found: {openai_api_key_path}"
            )
        load_dotenv(dotenv_path=openai_api_key_path)
        if "OPENAI_API_KEY" not in os.environ:
            raise ValueError("OpenAI API key not found in environment variables.")
        llm = LLM(model_name="gpt-4o-mini", model_type=ChatOpenAI, temperature=0.0)
        ticker_mood.set_llm(llm)
    ticker_mood.run()



import logging
import os
import urllib.parse
from datetime import datetime
from typing import List, Optional, Type

import ollama
from langchain_core.language_models import BaseChatModel
from pydantic import BaseModel, Field, model_validator

from tickermood.articles import News, PriceTargetNews, NewsSummary, Summary
from tickermood.database.crud import TickerMoodDb
from langchain_ollama import ChatOllama
from langchain_openai import ChatOpenAI

from tickermood.exceptions import InvalidLLMError
from tickermood.types import DatabaseConfig

logger = logging.getLogger(__name__)


class TickerSubject(BaseModel):
    symbol: str
    symbol_without_suffix: Optional[str] = Field(default=None)
    name: Optional[str] = None
    exchange: Optional[str] = None

    @model_validator(mode="after")
    def _validation(self) -> "TickerSubject":
        self.symbol_without_suffix = self.symbol.split(".")[0]
        return self

    def to_symbol_search(self) -> str:
        if self.symbol_without_suffix is None:
            raise ValueError("Symbol without suffix is not set.")
        if self.name:
            return urllib.parse.quote(
                f"{self.symbol_without_suffix.upper()}+{self.name.replace(' ', '+')}"
            )
        return self.symbol_without_suffix.upper()


class PriceTarget(BaseModel):
    high_price_target: Optional[float] = None
    low_price_target: Optional[float] = None
    fair_value: Optional[float] = None
    summary_price_target: Optional[str] = None


class Consensus(BaseModel):
    consensus: Optional[str] = None
    reason: Optional[str] = None


class Subject(TickerSubject, PriceTarget, Consensus):
    date: Optional[datetime] = Field(default_factory=datetime.now)
    news: List[News] = Field(default_factory=list)
    news_summary: List[NewsSummary] = Field(default_factory=list)
    summary: Optional[Summary] = None
    price_target_news: List[PriceTargetNews] = Field(default_factory=list)

    @model_validator(mode="after")
    def _validator(self) -> "Subject":
        if self.news_summary and len(self.news_summary) != len(self.news_summary):
            raise ValueError(
                "The length of news_summary must match the length of news."
            )
        if len(self.news) != len(set(self.news)):
            raise ValueError("News items must be unique.")
        return self

    def save(self, database_config: DatabaseConfig) -> None:
        TickerMoodDb(
            database_path=database_config.database_path,
            no_migration=database_config.no_migration,
        ).write(self)

    def load(self, database_config: DatabaseConfig) -> "Subject":
        db = TickerMoodDb(
            database_path=database_config.database_path,
            no_migration=database_config.no_migration,
        )
        return db.load(subject=self)

    def add_news_summary(self, content: str, origin: News) -> None:
        self.news_summary.append(
            NewsSummary(
                url=origin.url,
                content=content,
                source=origin.source,
                title=origin.title,
            )
        )

    def add_summary(self, content: str) -> None:
        self.summary = Summary(content=content)

    def combined_summary_news(self) -> str:
        return "####\n".join(n.content for n in self.news_summary if n.content)

    def combined_price_target_news(self) -> str:
        return "####\n".join(n.content for n in self.price_target_news if n.content)

    def add(self, object: BaseModel) -> None:
        for field in list(object.model_fields):
            setattr(self, field, getattr(object, field))

    def get_consensus_data(self) -> str:
        data = f"""Summary of the stock:
        {self.summary}
        {self.summary_price_target}
        """
        return data


def check_ollama_model(model_name: str) -> bool:
    try:
        model_list = ollama.list()
    except Exception as e:
        logger.error(e)
        return False
    if not any(model.model == model_name for model in model_list.models):
        logger.error(f"Ollama model '{model_name}' not found.")
        return False

    return True


def check_openai_model() -> bool:
    return "OPENAI_API_KEY" in os.environ


class LLM(BaseModel):
    model_type: Type[BaseChatModel]
    model_name: str
    temperature: float = 0.0

    @model_validator(mode="after")
    def _validator(self) -> "LLM":
        if (self.model_type == ChatOllama and check_ollama_model(self.model_name)) or (
            self.model_type == ChatOpenAI and check_openai_model()
        ):
            return self
        raise InvalidLLMError("Only Ollama and OpenAI models are supported.")

    def get_model(self) -> BaseChatModel:
        return self.model_type(model=self.model_name, temperature=self.temperature)


class LLMSubject(Subject, LLM):

    @classmethod
    def from_subject(cls, subject: Subject, llm: LLM) -> "LLMSubject":
        return cls.model_validate(subject.model_dump() | llm.model_dump())

    def get_next_article(self) -> Optional[News]:
        return next(
            (
                n
                for n in self.news
                if hash(n) not in {hash(s) for s in self.news_summary}
            ),
            None,
        )



import json
import logging
import re
from typing import get_args, Type

from langchain_core.messages import SystemMessage, HumanMessage
from langchain_core.output_parsers import JsonOutputParser
from langgraph.graph import StateGraph
from langgraph.graph.state import CompiledStateGraph
from pydantic import BaseModel

from tickermood.subject import Subject, LLMSubject, PriceTarget, Consensus
from tickermood.types import ConsensusType

logger = logging.getLogger(__name__)


def remove_tagged_text(text: str) -> str:
    pattern = r"<think>.*?</think>"
    return re.sub(pattern, "", text, flags=re.DOTALL).strip()


def get_json_schema(model: Type[BaseModel]) -> str:

    return json.dumps(model.model_json_schema().get("properties", {}), indent=4)


def parse_json_output(model_response: str, model: Type[BaseModel]) -> BaseModel:
    parser = JsonOutputParser()
    try:
        parsed_output = parser.parse(remove_tagged_text(model_response))
        return model.model_validate(parsed_output)
    except Exception as e:
        logger.error(f"Failed to parse model response into {model.__name__}: {e}")
        return model()


def summarize(state: LLMSubject) -> LLMSubject:
    llm = state.get_model()
    system_message = SystemMessage(
        "You are a helpful assistant that summarizes financial articles. "
        "Reasoning, thought process, or annotations like <think>. "
        "Only return the final summary in plain text. No tags, no notes, no process."
        "Only few sentences."
    )
    article = state.get_next_article()
    if article:
        human_message = HumanMessage(
            "Summarize the following. Only return the summary.:\n\n" + article.content
        )
        response = llm.invoke([system_message, human_message])
        state.add_news_summary(remove_tagged_text(str(response.content)), article)
    return state


def has_more_articles(state: LLMSubject) -> bool:
    return state.get_next_article() is not None


def reduce(state: LLMSubject) -> LLMSubject:
    llm = state.get_model()
    system_message = SystemMessage(
        "You are a helpful and smart financial assistant that can summarizes finance articles."
    )
    human_message = HumanMessage(
        "Provide a concise summary of these articles. Is this stock a GO/NO-GO/Cautious?:\n\n"
        + state.combined_summary_news()
    )
    response = llm.invoke([system_message, human_message])
    state.add_summary(remove_tagged_text(str(response.content)))
    return state


def price_target(state: LLMSubject) -> LLMSubject:
    llm = state.get_model()
    system_message = SystemMessage(
        "You are a helpful assistant that summarizes financial articles."
    )
    human_message = HumanMessage(
        f"""Extract the high and low price target and give a summary of the text. 
        Return the answer strictly in this JSON format:
{get_json_schema(PriceTarget)}:\n\n"""
        + state.combined_price_target_news()
    )
    response = llm.invoke([system_message, human_message])
    state.add(parse_json_output(str(response.content), PriceTarget))
    return state


def get_consensus(state: LLMSubject) -> LLMSubject:
    llm = state.get_model()

    system_message = SystemMessage(
        "You are a helpful assistant that summarizes financial articles."
    )
    human_message = HumanMessage(
        f"Summarize the analyst consensus for the stock with symbol {state.symbol} and return using the following "
        f"JSON format:{get_json_schema(Consensus)}.\n\n"
        f"Note that the consensus field must be one of the following {list(get_args(ConsensusType))}\n\n"
        + state.get_consensus_data()
    )
    response = llm.invoke([system_message, human_message])
    state.add(parse_json_output(str(response.content), Consensus))
    return state


def summarize_agent() -> CompiledStateGraph:
    graph = StateGraph(LLMSubject)

    graph.add_node("summarize", summarize)
    graph.add_node("reduce", reduce)
    graph.add_node("price_target", price_target)
    graph.add_node("get_consensus", get_consensus)
    graph.set_entry_point("summarize")
    graph.add_conditional_edges(
        "summarize", has_more_articles, {True: "summarize", False: "reduce"}
    )
    graph.add_edge("reduce", "price_target")
    graph.add_edge("price_target", "get_consensus")

    graph.set_finish_point("get_consensus")

    return graph.compile()


def invoke_summarize_agent(subject: LLMSubject) -> Subject:

    graph = summarize_agent()
    subject = graph.invoke(subject)  # type: ignore
    return Subject.model_validate(subject)



import os
from pathlib import Path
from typing import Optional

from alembic import command
from alembic.config import Config
from sqlalchemy import create_engine


from tickermood.database.settings import DATABASE_URL


def upgrade(database_url: str, no_migration: Optional[bool] = False) -> None:
    root_folder = Path(__file__).parents[1]
    os.environ.update({"DATABASE_URL": database_url})
    alembic_cfg = Config(root_folder / "alembic" / "alembic.ini")
    alembic_cfg.set_main_option("script_location", str(root_folder / "alembic"))
    if no_migration:
        from tickermood.database.schemas import SubjectORM

        engine = create_engine(database_url, echo=True)
        SubjectORM.__table__.create(engine, checkfirst=True)  # type: ignore
    else:
        command.upgrade(alembic_cfg, "head")


if __name__ == "__main__":
    upgrade(DATABASE_URL)



from functools import cached_property
from pathlib import Path
from typing import Any, TYPE_CHECKING

from pydantic import BaseModel, ConfigDict
from sqlalchemy import Engine, create_engine, insert
from sqlmodel import Session, select


from tickermood.database.scripts.upgrade import upgrade

if TYPE_CHECKING:
    from tickermood.subject import Subject


class TickerMoodDb(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    database_path: Path
    no_migration: bool = False

    @cached_property
    def _engine(self) -> Engine:
        database_url = f"sqlite:///{Path(self.database_path)}"
        upgrade(database_url, no_migration=self.no_migration)
        engine = create_engine(database_url)
        return engine

    def model_post_init(self, __context: Any) -> None:
        self._engine  # noqa: B018

    def write(self, subject: "Subject") -> None:
        from tickermood.database.schemas import SubjectORM

        with Session(self._engine) as session:
            stmt = (
                insert(SubjectORM)
                .prefix_with("OR REPLACE")
                .values([subject.model_dump()])
            )
            session.exec(stmt)  # type: ignore
            session.commit()

    def load(self, subject: "Subject") -> "Subject":
        from tickermood.database.schemas import SubjectORM
        from tickermood.subject import Subject

        with Session(self._engine) as session:
            stmt = (
                select(SubjectORM)
                .where(SubjectORM.symbol == subject.symbol)
                .order_by(SubjectORM.date.desc())  # type: ignore
                .limit(1)
            )
            result = session.exec(stmt).first()
            if result is None:
                raise ValueError(f"No data found for symbol: {subject.symbol}")
            return Subject.model_validate(result.model_dump())



from pathlib import Path
from typing import Literal

from pydantic import BaseModel, Field

SourceName = Literal["Investing", "Marketwatch", "Yahoo"]
ConsensusType = Literal[
    "Strong Buy", "Buy", "Cautious Buy", "Hold", "Cautious Sell", "Sell", "Strong Sell"
]


class DatabaseConfig(BaseModel):
    database_path: Path = Field(default=Path.cwd() / "tickermood.db")
    no_migration: bool = False



from datetime import datetime
from typing import Optional, List, Any

from sqlalchemy import JSON, Column
from sqlmodel import SQLModel, Field

from tickermood.subject import Subject


class BaseTable(SQLModel): ...


class SubjectORM(BaseTable, Subject, table=True):
    __tablename__ = "subject"
    date: datetime = Field(primary_key=True, index=True)
    symbol: str = Field(primary_key=True, index=True)
    consensus: Optional[str] = Field(default=None, index=True)
    news: Optional[List[Any]] = Field(default=None, sa_column=Column(JSON))  # type: ignore
    news_summary: Optional[List[Any]] = Field(default=None, sa_column=Column(JSON))  # type: ignore
    summary: Optional[List[Any]] = Field(default=None, sa_column=Column(JSON))  # type: ignore
    price_target_news: Optional[List[Any]] = Field(default=None, sa_column=Column(JSON))  # type: ignore
